"""A Slack bot using the real-time messaging API."""

from itertools import count
import json
import logging
from random import randint
from textwrap import dedent

from aiohttp import MsgType, ws_connect

from .slack_api import SlackApiError, SlackBotApi

logger = logging.getLogger(__name__)


class SlackBot:
    """Base class Slack bot.

        Arguments:
          id_ (:py:class:`str`): The bot's Slack ID.
          user (:py:class:`str`): The bot's friendly name.
          api (SlackApi): The Slack API wrapper.

        Attributes:
          address_as (:py:class:`str`): The text that appears at the
            start of messages address to this bot (e.g.
            ``'<@user>: '``).
          full_name (:py:class:`str`): The name of the bot as it
            appears in messages about the bot (e.g. ``'<@user>'``).

        """

    API_AUTH_ENDPOINT = 'auth.test'
    """Test endpoint for API authorisation."""

    INSTRUCTIONS = dedent("""
    These are the default instructions for an aSlack bot.

    Override these as appropriate for your specific needs.
    """)

    MESSAGE_FILTERS = {}
    """Default filters for incoming messages."""

    RTM_HANDSHAKE = {'type': 'hello'}
    """Expected handshake message from RTM API."""

    RTM_START_ENDPOINT = 'rtm.start'
    """Start endpoint for real-time messaging."""

    def __init__(self, id_, user, api):
        self.id_ = id_
        self.user = user
        self.api = api
        self.full_name = '<@{}>'.format(id_)
        self.address_as = '{}: '.format(self.full_name)
        self._msg_ids = count(randint(1, 1000))

    async def get_socket_url(self):
        """Get the WebSocket URL for the RTM session.

        Warning:
          The URL expires if the session is not joined within 30
          seconds of the API call to the start endpoint.

        Returns:
          :py:class:`str`: The socket URL.

        """
        data = await self.api.execute_method(self.RTM_START_ENDPOINT)
        return data['url']

    def handle_message(self, message, filters):
        """Handle an incoming message appropriately.

        Arguments:
          message (:py:class:`aiohttp.websocket.Message`): The incoming
            message to handle.
          filters (:py:class:`dict`): The filters to apply to incoming
            messages.

        Returns:
          :py:class:`str`: The response to send as a result.

        """
        data = self._unpack_message(message)
        logger.debug(data)
        if data.get('type') == 'error':
            raise SlackApiError(
                data.get('error', {}).get('msg', str(data))
            )
        elif (self.message_is_to_me(data) and
              data['text'][len(self.address_as):].strip() == '?'):
            return self._format_message(
                channel=data['channel'],
                text=self._instruction_list(filters),
            )
        for filter_, dispatch in filters.items():
            if filter_(data):
                logger.debug('Response triggered')
                return self._format_message(**dispatch(data))

    async def join_rtm(self, filters=None):
        """Join the real-time messaging service.

        Arguments:
          filters (:py:class:`dict`, optional): Dictionary mapping
            message filters to the functions they should dispatch to.
            Use a :py:class:`collections.OrderedDict` if precedence is
            important; only one filter, the first match, will be
            applied to each message.

        """
        if filters is None:
            filters = self.MESSAGE_FILTERS
        url = await self.get_socket_url()
        logger.debug('Connecting to {!r}'.format(url))
        async with ws_connect(url) as socket:
            first_msg = await socket.receive()
            self._validate_first_message(first_msg)
            async for message in socket:
                if message.tp == MsgType.text:
                    result = self.handle_message(message, filters)
                    if result is not None:
                        socket.send_str(result)
                elif message.tp in (MsgType.closed, MsgType.error):
                    if not socket.closed:
                        await socket.close()
                    break
        logger.info('Left real-time messaging.')

    def message_mentions_me(self, data):
        """If you send a message that mentions the bot's name"""
        return (data.get('type') == 'message' and
                self.full_name in data.get('text', ''))

    def message_is_to_me(self, data):
        """If you send a message starting '@username: ' to the bot"""
        return (data.get('type') == 'message' and
                data.get('text', '').startswith(self.address_as))

    @classmethod
    async def from_api_token(cls, token):
        """Create a new instance from the API token.

        Arguments:
          token (:py:class:`str`): The bot's API token.

        Returns:
          SlackBot: The new instance.

        """
        api = SlackBotApi(token)
        data = await api.execute_method(cls.API_AUTH_ENDPOINT)
        return cls(data['user_id'], data['user'], api)

    def _format_message(self, channel, text):
        """Format an outoging message for transmission.

        Note:
          Adds the message type (``'message'``) and incremental ID.

        Arguments:
          channel (:py:class:`str`): The channel to send to.
          text (:py:class:`str`): The message text to send.

        Returns:
          str: The JSON string of the message.

        """
        payload = {'type': 'message', 'id': next(self._msg_ids)}
        payload.update(channel=channel, text=text)
        return json.dumps(payload)

    @classmethod
    def _instruction_list(cls, filters):
        """Generates the instructions for a bot and its filters.

        Note:
          The guidance for each filter is generated by combining the
          docstrings of the predicate filter and resulting dispatch
          function with a single space between. The class's
          :py:attr:`INSTRUCTIONS` and the default help command are
          added.

        Arguments:
          filters (:py:class:`dict`): The filters to apply to incoming
            messages.

        Returns:
          :py:class:`str`: The bot's instructions.

        """
        return '\n\n'.join([
            cls.INSTRUCTIONS.strip(),
            '*Supported methods:*',
            'If you send "@username: ?" to the bot it replies with these '
            'instructions.',
        ] + [
            ' '.join((filter_.__doc__, dispatch.__doc__))
            for filter_, dispatch in filters.items()
        ])

    @classmethod
    def _validate_first_message(cls, msg):
        """Check the first message matches the expected handshake.

        Arguments:
          msg (:py:class:`aiohttp.Message`): The message to validate.

        Raises:
          SlackApiError: If the data doesn't match the handshake.

        """
        data = cls._unpack_message(msg)
        logger.debug(data)
        if data != cls.RTM_HANDSHAKE:
            raise SlackApiError('Unexpected response: {!r}'.format(data))
        logger.info('Joined real-time messaging.')

    @staticmethod
    def _unpack_message(msg):
        """Unpack the data from the message.

        Arguments:
          msg (:py:class:`aiohttp.Message`): The message to unpack.

        Returns:
          :py:class:`dict`: The loaded data.

        Raises:
          :py:class:`AttributeError`: If there is no data attribute.
          :py:class:`json.JSONDecodeError`: If the data isn't valid
            JSON.

        """
        return json.loads(msg.data)



"""Utility functionality."""

import os

from aiohttp import web_exceptions

API_TOKEN_ENV = 'SLACK_API_TOKEN'
"""The environment variable to store the user's API token in."""


class FriendlyError(Exception):
    """Exception with friendlier error messages.

      Notes:
        The ``err_msg`` is resolved in :py:data:`EXPECTED_ERRORS`,
        or passed through as-is if not found there.

      Arguments:
        err_msg (:py:class:`str`): The error message to attempt to
          resolve.
        *args (:py:class:`tuple`): Any additional positional arguments.

    """

    EXPECTED_ERRORS = {}
    """Friendly messages for expected errors."""

    def __init__(self, err_msg, *args):
        super().__init__(self.EXPECTED_ERRORS.get(err_msg, err_msg), *args)


def raise_for_status(response):
    """Raise an appropriate error for a given response.

    Arguments:
      response (:py:class:`aiohttp.ClientResponse`): The API response.

    Raises:
      :py:class:`aiohttp.web_exceptions.HTTPException`: The appropriate
        error for the response's status.

    """
    for err_name in web_exceptions.__all__:
        err = getattr(web_exceptions, err_name)
        if err.status_code == response.status:
            payload = dict(
                headers=response.headers,
                reason=response.reason,
            )
            if issubclass(err, web_exceptions._HTTPMove):
                raise err(response.headers['Location'], **payload)
            raise err(**payload)


def get_api_token():
    """Allow the user to enter their API token.

    Note:
      The token is added to the environment using the variable defined
      in :py:attr:`API_TOKEN_ENV`.

    Returns:
      :py:class:`str`: The user's API token.

    """
    token = os.getenv(API_TOKEN_ENV)
    if token:
        return token
    template = ('Enter your API token (this will be stored '
                'as {} for future use): ').format(API_TOKEN_ENV)
    token = input(template)
    os.environ[API_TOKEN_ENV] = token
    return token



"""Access to the base Slack Web API."""

import logging

import aiohttp

from .utils import get_api_token, FriendlyError, raise_for_status

logger = logging.getLogger(__name__)


class SlackApiError(FriendlyError):
    """Wrapper exception for error messages in the response JSON."""

    EXPECTED_ERRORS = {
        'account_inactive': 'Authentication token is for a deleted user or '
                            'team.',
        'invalid_auth': 'Invalid authentication token.',
        'migration_in_progress': 'Team is being migrated between servers.',
        'not_authed': "No authentication token provided.",
    }
    """Friendly messages for expected Slack API errors."""


class SlackApi:
    """Class to handle interaction with Slack's API.

    Arguments:
      :py:class:`str`: The user's API token.

    """

    API_BASE_URL = 'https://slack.com/api'
    """The base URL for Slack API calls."""

    API_METHODS = {
        'api': {'test': {'Checks API calling code.'}},
        'auth': {'test': 'Checks authentication & identity.'},
        'channels': {
            'archive': 'Archives a channel.',
            'create': 'Creates a channel.',
            'history': 'Fetches history of messages and events from a channel.',
            'info': 'Gets information about a channel.',
            'invite': 'Invites a user to a channel.',
            'join': 'Joins a channel, creating it if needed.',
            'kick': 'Removes a user from a channel.',
            'leave': 'Leaves a channel.',
            'list': 'Lists all channels in a Slack team.',
            'mark': 'Sets the read cursor in a channel.',
            'rename': 'Renames a channel.',
            'setPurpose': 'Sets the purpose for a channel.',
            'setTopic': 'Sets the topic for a channel.',
            'unarchive': 'Unarchives a channel.',
        },
        'chat': {
            'delete': 'Deletes a message.',
            'postMessage': 'Sends a message to a channel.',
            'update': 'Updates a message.'
        },
        'emoji': {'list': '	Lists custom emoji for a team.'},
        'files': {
            'delete': 'Deletes a file.',
            'info': 'Gets information about a team file.',
            'list': 'Lists & filters team files.',
            'upload': 'Uploads or creates a file.'
        },
        'groups': {
            'archive': 'Archives a private channel.',
            'close': 'Closes a private channel.',
            'create': 'Creates a private private channel.',
            'createChild': 'Clones and archives a private channel.',
            'history': 'Fetches history of messages and events from a private '
                       'channel.',
            'info': 'Gets information about a private channel.',
            'invite': 'Invites a user to a private channel.',
            'kick': 'Removes a user from a private channel.',
            'leave': 'Leaves a private channel.',
            'list': 'Lists private channels that the calling user has access '
                    'to.',
            'mark': 'Sets the read cursor in a private channel.',
            'open': 'Opens a private channel.',
            'rename': 'Renames a private channel.',
            'setPurpose': 'Sets the purpose for a private channel.',
            'setTopic': 'Sets the topic for a private channel.',
            'unarchive': 'Unarchives a private channel.',
        },
        'im': {
            'close': 'Close a direct message channel.',
            'history': 'Fetches history of messages and events from direct '
                       'message channel.',
            'list': 'Lists direct message channels for the calling user.',
            'mark': 'Sets the read cursor in a direct message channel.',
            'open': 'Opens a direct message channel.',
        },
        'mpim': {
            'close': 'Closes a multiparty direct message channel.',
            'history': 'Fetches history of messages and events from a '
                       'multiparty direct message.',
            'list': 'Lists multiparty direct message channels for the calling '
                    'user.',
            'mark': 'Sets the read cursor in a multiparty direct message '
                    'channel.',
            'open': 'This method opens a multiparty direct message.',
        },
        'oauth': {
            'access': 'Exchanges a temporary OAuth code for an API token.'
        },
        'pins': {
            'add': 'Pins an item to a channel.',
            'list': 'Lists items pinned to a channel.',
            'remove': 'Un-pins an item from a channel.',
        },
        'reactions': {
            'add': 'Adds a reaction to an item.',
            'get': 'Gets reactions for an item.',
            'list': 'Lists reactions made by a user.',
            'remove': 'Removes a reaction from an item.',
        },
        'rtm': {'start': 'Starts a Real Time Messaging session.'},
        'search': {
            'all': 'Searches for messages and files matching a query.',
            'files': 'Searches for files matching a query.',
            'messages': 'Searches for messages matching a query.',
        },
        'stars': {
            'add': 'Adds a star to an item.',
            'list': 'Lists stars for a user.',
            'remove': 'Removes a star from an item.',
        },
        'team': {
            'accessLogs': 'Gets the access logs for the current team.',
            'info': 'Gets information about the current team.',
            'integrationLogs': 'Gets the integration logs for the current '
                               'team.',
        },
        'usergroups': {
            'create': 'Create a user group.',
            'disable': 'Disable an existing user group.',
            'enable': 'Enable a user group.',
            'list': 'List all user groups for a team.',
            'update': 'Update an existing user group',
            'users': {
                'list': 'List all users in a user group',
                'update': '	Update the list of users for a user group.',
            },
        },
        'users': {
            'getPresence': 'Gets user presence information.',
            'info': 'Gets information about a user.',
            'list': 'Lists all users in a Slack team.',
            'setActive': 'Marks a user as active.',
            'setPresence': 'Manually sets user presence.',
        },
    }
    """The API methods defined by Slack."""

    def __init__(self, token=None):
        if token is None:
            token = get_api_token()
        self.token = token

    async def execute_method(self, method, **params):
        """Execute a specified Slack Web API method.

        Note:
          The API token is added automatically by this method.

        Arguments:
          method (:py:class:`str`): The name of the method.
          **params (:py:class:`dict`): Any additional parameters
            required.

        Returns:
          :py:class:`dict`: The JSON data from the response.

        Raises:
          :py:class:`aiohttp.web_exceptions.HTTPException`: If the HTTP
            request returns a code other than 200 (OK).
          SlackApiError: If the Slack API is reached but the response
           contains an error message.

        """
        url = self._create_url(method)
        params = params.copy()
        params['token'] = self.token
        logger.info('Executing method {!r}'.format(method))
        logger.debug('...with params {!r}'.format(params))
        response = await aiohttp.get(url, params=params)
        logger.info('Status: {}'.format(response.status))
        if response.status == 200:
            json = await response.json()
            logger.debug('...with JSON {!r}'.format(json))
            if json.get('ok'):
                return json
            raise SlackApiError(json['error'])
        else:
            raise_for_status(response)

    @classmethod
    def _create_url(cls, method):
        """Create the full API URL for a given method.

        Arguments:
          method (:py:class:`str`): The name of the method.

        Returns:
          :py:class:`str`: The full API URL.

        Raises:
          SlackApiError: If the method is unknown.

        """
        if not cls._method_exists(method):
            raise SlackApiError('The {!r} method is unknown.'.format(method))
        return '/'.join((cls.API_BASE_URL, method))

    @classmethod
    def _method_exists(cls, method):
        """Whether a given method exists in the known API.

        Note:
          :py:class:`SlackApi` and its subclasses provide a dictionary
          of :py:attr:`API_METHODS`, a class attribute defining the
          known API.

        Arguments:
          method (:py:class:`str`): The name of the method.

        Returns:
          :py:class:`bool`: Whether the method is in the known API.

        """
        methods = cls.API_METHODS
        for key in method.split('.'):
            methods = methods.get(key)
            if methods is None:
                break
        if isinstance(methods, str):
            logger.debug('{!r}: {!r}'.format(method, methods))
            return True
        return False


class SlackBotApi(SlackApi):
    """API accessible to Slack custom bots."""

    API_METHODS = {
        'api': {'test': {'Checks API calling code.'}},
        'auth': {'test': 'Checks authentication & identity.'},
        'channels': {
            'history': 'Fetches history of messages and events from a channel.',
            'info': 'Gets information about a channel.',
            'list': 'Lists all channels in a Slack team.',
            'mark': 'Sets the read cursor in a channel.',
            'setPurpose': 'Sets the purpose for a channel.',
            'setTopic': 'Sets the topic for a channel.',
        },
        'chat': {
            'delete': 'Deletes a message.',
            'postMessage': 'Sends a message to a channel.',
            'update': 'Updates a message.'
        },
        'emoji': {'list': '	Lists custom emoji for a team.'},
        'files': {
            'delete': 'Deletes a file.',
            'upload': 'Uploads or creates a file.'
        },
        'groups': {
            'close': 'Closes a private channel.',
            'history': 'Fetches history of messages and events from a private '
                       'channel.',
            'info': 'Gets information about a private channel.',
            'list': 'Lists private channels that the calling user has access '
                    'to.',
            'mark': 'Sets the read cursor in a private channel.',
            'open': 'Opens a private channel.',
            'rename': 'Renames a private channel.',
            'setPurpose': 'Sets the purpose for a private channel.',
            'setTopic': 'Sets the topic for a private channel.',
        },
        'im': {
            'close': 'Close a direct message channel.',
            'history': 'Fetches history of messages and events from direct '
                       'message channel.',
            'list': 'Lists direct message channels for the calling user.',
            'mark': 'Sets the read cursor in a direct message channel.',
            'open': 'Opens a direct message channel.',
        },
        'mpim': {
            'close': 'Closes a multiparty direct message channel.',
            'history': 'Fetches history of messages and events from a '
                       'multiparty direct message.',
            'list': 'Lists multiparty direct message channels for the calling '
                    'user.',
            'mark': 'Sets the read cursor in a multiparty direct message '
                    'channel.',
            'open': 'This method opens a multiparty direct message.',
        },
        'oauth': {
            'access': 'Exchanges a temporary OAuth code for an API token.'
        },
        'pins': {
            'add': 'Pins an item to a channel.',
            'remove': 'Un-pins an item from a channel.',
        },
        'reactions': {
            'add': 'Adds a reaction to an item.',
            'get': 'Gets reactions for an item.',
            'list': 'Lists reactions made by a user.',
            'remove': 'Removes a reaction from an item.',
        },
        'rtm': {'start': 'Starts a Real Time Messaging session.'},
        'stars': {
            'add': 'Adds a star to an item.',
            'remove': 'Removes a star from an item.',
        },
        'team': {'info': 'Gets information about the current team.'},
        'users': {
            'getPresence': 'Gets user presence information.',
            'info': 'Gets information about a user.',
            'list': 'Lists all users in a Slack team.',
            'setActive': 'Marks a user as active.',
            'setPresence': 'Manually sets user presence.',
        },
    }
    """The API methods defined by Slack."""


class SlackAppBotApi(SlackApi):
    """API accessible to Slack app bots."""

    API_METHODS = {
        'api': {'test': {'Checks API calling code.'}},
        'auth': {'test': 'Checks authentication & identity.'},
        'channels': {
            'info': 'Gets information about a channel.',
            'list': 'Lists all channels in a Slack team.',
        },
        'chat': {
            'delete': 'Deletes a message.',
            'postMessage': 'Sends a message to a channel.',
            'update': 'Updates a message.'
        },
        'files': {
            'delete': 'Deletes a file.',
            'upload': 'Uploads or creates a file.'
        },
        'groups': {
            'info': 'Gets information about a private channel.',
            'list': 'Lists private channels that the calling user has access '
                    'to.',
        },
        'im': {
            'close': 'Close a direct message channel.',
            'history': 'Fetches history of messages and events from direct '
                       'message channel.',
            'list': 'Lists direct message channels for the calling user.',
            'mark': 'Sets the read cursor in a direct message channel.',
            'open': 'Opens a direct message channel.',
        },
        'mpim': {
            'close': 'Closes a multiparty direct message channel.',
            'history': 'Fetches history of messages and events from a '
                       'multiparty direct message.',
            'list': 'Lists multiparty direct message channels for the calling '
                    'user.',
            'mark': 'Sets the read cursor in a multiparty direct message '
                    'channel.',
            'open': 'This method opens a multiparty direct message.',
        },
        'oauth': {
            'access': 'Exchanges a temporary OAuth code for an API token.'
        },
        'pins': {
            'add': 'Pins an item to a channel.',
            'remove': 'Un-pins an item from a channel.',
        },
        'reactions': {
            'add': 'Adds a reaction to an item.',
            'get': 'Gets reactions for an item.',
            'list': 'Lists reactions made by a user.',
            'remove': 'Removes a reaction from an item.',
        },
        'rtm': {'start': 'Starts a Real Time Messaging session.'},
        'stars': {
            'add': 'Adds a star to an item.',
            'remove': 'Removes a star from an item.',
        },
        'users': {
            'getPresence': 'Gets user presence information.',
            'info': 'Gets information about a user.',
            'list': 'Lists all users in a Slack team.',
            'setActive': 'Marks a user as active.',
            'setPresence': 'Manually sets user presence.',
        },
    }
    """The API methods defined by Slack."""
